import torch
import torchvision.transforms as transforms
import matplotlib.pyplot as plt
from PIL import Image
import cv2
import os
from blind_deconv import blind_deconv
from ringing_artifacts_removal import ringing_artifacts_removal
from misc import visualize_rgb ,visualize_image, gray_image, process_image,PSNR
from metrics import psnr
import time
import sys


def main():
    
        # if(i!=15):
        #     continue
        # plt.figure(figsize=(5, 5))
        # Specify your input image file path
        # image_path = 'images/blurry1_8.png' 41
    kernel_list = [87, 93, 99, 105, 111, 117, 123, 129, 135, 141]
    prefix = 'manmade'
    for coco in range(10):
        for img_idx in range(5):
            for ker_idx in range(2,4):
                try:
                    
                    # if img_idx+1 == 2 and ker_idx+1 == 4:
                    #     continue
                    # if img_idx+1 == 3 and ker_idx+1 == 1:
                    #     continue
                    image_path = f'images/Lai/{prefix}_0{img_idx+1}_kernel_0{ker_idx+1}.png'

                    # print(sys.argv[1,1])
                    kernel_size = 77
                    opts = {
                        'prescale': 1,   # Downsampling
                        'xk_iter': 5,    # Iterations
                        'gamma_correct': 1.0,
                        'k_thresh': 20,
                        'kernel_size':kernel_list[coco] ,
                    }
                    # print(sys.argv[1])
                    lambda_dark = 4e-3


                    #Experimenting with lambda_dark set to 0
                    #*************************************EXPERIMENT !*************************************
                    # lambda_ftr = 3e-4
                    # # lambda_ftr = 1000
                    # lambda_dark = 0
                    # # lambda_grad = 3.87e-4
                    # # lambda_grad = 4e-3
                    # lambda_grad = 4e-3


                    # lambda_tv = 0.005
                    # # lambda_l0 = 5e-4
                    # lambda_l0 = 5e-4
                    # weight_ring = 1

                    # ********************************* EXPERIMENT 2 *************************************

                    lambda_dark = 4e-3
                    #Experimenting with lambda_dark set to 0
                    # lambda_ftr = 3.5e-4 # holy grail
                    
                    lambda_ftr = 3.5e-4
                    lambda_dark = 0
                    lambda_grad = 5e-3


                    lambda_tv = 1e-3
                    lambda_l0 = 2e-4
                    # # lambda_l0 = list[i]
                    
                    weight_ring = 0

                    results_dir = 'results'
                    os.makedirs(results_dir, exist_ok=True)

                    # Load the image
                    image = cv2.imread(image_path)

                    # Set parameters
                    
                    is_select = False  # Set to True if you want to select a specific area for deblurring

                    if is_select:
                        # Allow the user to select a specific area for deblurring (not implemented in this example)
                        pass
                    else:
                        inpt = Image.open(image_path)
                        yg = gray_image(inpt)

                        # print(yg[0:5,0:5])
                    # Perform blind deconvolution
                    start_time = time.time()
                    kernel, interim_latent = blind_deconv(yg, lambda_ftr,lambda_dark, lambda_grad, opts)
                    end_time = time.time()
                    print(f"Time taken: {end_time-start_time} seconds")
                    # plt.figure(figsize=(5, 5))
                    # plt.imshow(kernel, cmap='gray')
                    # plt.title('Estimated Kernel')
                    # plt.show()
                    # Perform non-blind deconvolution
                    saturation = 0  # Set this to 1 if the image is saturated
                    if not saturation:
                        # Apply TV-L2 denoising method
                        y = process_image(Image.open(image_path))
                        y = y.permute(1,2,0)
                        y = y/255.0
                        # print(y[0:5,0:5,0])
                        # print(y.shape)
                        Latent = ringing_artifacts_removal(y, kernel, lambda_tv, lambda_l0, weight_ring)
                    else:
                        # Apply Whyte's deconvolution method
                        # Latent = whyte_deconv(yg, kernel)
                        pass
                    # print(Latent.shape)
                    # print(Latent.max())
                    # Latent = Latent/255.0
                    # print(Latent[0:5,0:5,0])
                    
                    #save the Latent matrix as a JPG image in the results folder
                    Latent[Latent>1.0] = 1.0
                    Latent[Latent<0.0] = 0.0
                    # if Latent.shape[2] == 1:
                    #     visualize_image(Latent)
                    # else:
                    #     visualize_rgb(Latent)
                    Latent = Latent.squeeze()
                    Latent = Latent*255.0
                    Latent = Latent.numpy()
                    Latent = Latent.astype('uint8')
                    Latent = Image.fromarray(Latent)
                    Latent.save(os.path.join(results_dir, f'{prefix}_0{img_idx+1}_kernel_0{ker_idx+1}_{coco}.png'))

                    kmn = kernel.min()
                    kmx = kernel.max()
                    kernel = (kernel - kmn)/(kmx - kmn)
                    kernel = kernel*255.0
                    kernel = kernel.numpy()
                    kernel = kernel.astype('uint8')
                    kernel = Image.fromarray(kernel)
                    kernel.save(os.path.join(results_dir, f'{prefix}_0{img_idx+1}_kernel_0{ker_idx+1}_kernel_{coco}.png'))    
                except Exception as e:
                    print(e)
                    continue
    # Lmx = Latent.max()
    # Lmn = Latent.min()
    # Latent = (Latent - Lmn)/(Lmx - Lmn)
    # visualize_rgb(Latent)
    # Display the results
    # plt.figure(figsize=(12, 6))
    # plt.subplot(131)
    # plt.imshow(kernel, cmap='gray')
    # plt.title('Estimated Kernel')
    # plt.subplot(132)
    # plt.imshow(interim_latent, cmap='gray')
    # plt.title('Interim Latent Image')
    # plt.subplot(133)
    # plt.imshow(Latent, cmap='gray')
    # plt.title('Deblurred Image')

    # # Save the results
    # kernel_image = Image.fromarray((kernel * 255).astype('uint8'))
    # latent_image = Image.fromarray((Latent * 255).astype('uint8'))
    # interim_image = Image.fromarray((interim_latent * 255).astype('uint8'))

    # kernel_image.save(os.path.join(results_dir, 'kernel.png'))
    # latent_image.save(os.path.join(results_dir, 'result.png'))
    # interim_image.save(os.path.join(results_dir, 'interim_result.png'))

if __name__ == "__main__":
    main()